"""
This module contains a set of pywr._core.Parameter subclasses for defining control curve based parameters.
"""
from ._control_curves import (PiecewiseLinearControlCurve, ControlCurveParameter, BaseControlCurveParameter,
    ControlCurveInterpolatedParameter, ControlCurveIndexParameter)



import numpy as np
import inspyred
from . import BaseOptimisationWrapper


class InspyredOptimisationWrapper(BaseOptimisationWrapper):
    """ A wrapper to enable optimisation using inspyred.

    A generator, bounder and evaluator method are provided to use with the inspyred algorithms.
    """
    def generator(self, random, args):

        values = []
        for var in self.model_variables:
            lower = []
            upper = []
            if var.double_size > 0:
                lower.append(var.get_double_lower_bounds())
                upper.append(var.get_double_upper_bounds())

            if var.integer_size > 0:
                lower.append(var.get_integer_lower_bounds())
                upper.append(var.get_integer_upper_bounds())

            lower = np.concatenate(lower)
            upper = np.concatenate(upper)

            if len(lower) != len(upper):
                raise ValueError('Upper and lower bounds are different lengths. Malformed bound data from Parameter:'
                                 ' "{}"'.format(var.name))

            for i in range(len(lower)):
                values.append(random.uniform(lower[i], upper[i]))

        return values

    def evaluator(self, candidates, args):
        fitness = []
        for candidate in candidates:
            for ivar, var in enumerate(self.model_variables):
                l, u = self.model_variable_map[ivar], self.model_variable_map[ivar+1]
                if var.double_size > 0:
                    dj = slice(l, l + var.double_size)
                    var.set_double_variables(np.array(candidate[dj]))

                if var.integer_size > 0:
                    ij = slice(l + var.double_size, u)
                    # Convert integer variables to ints
                    ints = np.rint(np.array(candidate[ij])).astype(np.int)
                    var.set_integer_variables(ints)

            self.model.reset()
            run_stats = self.model.run()

            fitness.append(inspyred.ec.emo.Pareto([r.aggregated_value() for r in self.model_objectives]))
        return fitness

    def bounder(self, candidate, args):
        for ivar, var in enumerate(self.model_variables):
            lower = []
            upper = []
            if var.double_size > 0:
                lower.append(var.get_double_lower_bounds())
                upper.append(var.get_double_upper_bounds())

            if var.integer_size > 0:
                lower.append(var.get_integer_lower_bounds())
                upper.append(var.get_integer_upper_bounds())

            lower = np.concatenate(lower)
            upper = np.concatenate(upper)

            j = slice(self.model_variable_map[ivar], self.model_variable_map[ivar+1])
            candidate[j] = np.minimum(upper, np.maximum(lower, candidate[j]))
        return candidate



import numpy as np
from . import BaseOptimisationWrapper

import logging
logger = logging.getLogger(__name__)


class PygmoWrapper(BaseOptimisationWrapper):

    def fitness(self, solution):
        logger.info('Evaluating solution ...')

        for ivar, var in enumerate(self.model_variables):
            j = slice(self.model_variable_map[ivar], self.model_variable_map[ivar+1])
            var.set_double_variables(np.array(solution[j]).copy())

        self.model.reset()
        run_stats = self.model.run()

        objectives = []
        for r in self.model_objectives:
            sign = 1.0 if r.is_objective == 'minimise' else -1.0
            value = r.aggregated_value()
            objectives.append(sign*value)

        constraints = [r.aggregated_value() for r in self.model_constraints]

        # Return values to the solution
        logger.info('Evaluation complete!')
        return objectives + constraints

    def get_bounds(self):
        """ Return the variable bounds. """
        lower = []
        upper = []
        for var in self.model_variables:

            if var.double_size > 0:
                lower.append(var.get_double_lower_bounds())
                upper.append(var.get_double_upper_bounds())

            if var.integer_size > 0:
                lower.append(var.get_integer_lower_bounds())
                upper.append(var.get_integer_upper_bounds())

        lower = np.concatenate(lower)
        upper = np.concatenate(upper)

        if len(lower) != len(upper):
            raise ValueError(
                'Upper and lower bounds are different lengths. Malformed bound data from Parameter:'
                ' "{}"'.format(var.name))

        return lower, upper

    def get_nobj(self):
        return len(self.model_objectives)

    def get_nec(self):
        return len(self.model_constraints)


import pandas
import numpy as np
import datetime
from pywr import _core

# Constants
SECONDS_IN_DAY = 3600 * 24


class Timestepper(object):
    def __init__(self, start="2015-01-01", end="2015-12-31", delta=1):
        self.start = start
        self.end = end
        self.delta = delta
        self._last_length = None
        self._periods = None
        self._deltas = None
        self._current = None
        self._next = None
        self.setup()
        self.reset()
        self._dirty = True

    def __iter__(self, ):
        return self

    def __len__(self, ):
        return len(self._periods)

    @property
    def dirty(self):
        return self._dirty

    def setup(self):
        periods = self.datetime_index

        # Compute length of each period
        deltas = periods.to_timestamp(how='e') - periods.to_timestamp(how='s')
        # Round to nearest second
        deltas = np.round(deltas.total_seconds())
        # Convert to days
        deltas = deltas / SECONDS_IN_DAY
        self._periods = periods
        self._deltas = deltas
        self.reset()
        self._dirty = False

    def reset(self, start=None):
        """ Reset the timestepper

        If start is None it resets to the original self.start, otherwise
        start is used as the new starting point.
        """
        self._current = None
        current_length = len(self)

        if start is None:
            current_index = 0
        else:
            # Determine which period the start time is within
            for index, period in enumerate(self._periods):
                if period.start_time <= start < period.end_time:
                    current_index = index
                    break
            else:
                raise ValueError('New starting position is outside the range of the model timesteps.')

        self._next = _core.Timestep(self._periods[current_index], current_index, self._deltas[current_index])
        length_changed = self._last_length != current_length
        self._last_length = current_length
        return length_changed

    def __next__(self, ):
        return self.next()

    def next(self, ):
        self._current = current = self._next

        if current.index >= len(self._periods):
            raise StopIteration()

        # Increment to next timestep
        next_index = current.index + 1
        if next_index >= len(self._periods):
            # The final time-step is one offset beyond the end of the model.
            # Here we compute its delta and create the object. 
            final_period = current.period + self.offset
            delta = final_period.end_time - final_period.start_time
            delta = np.round(delta.total_seconds())
            delta = delta / SECONDS_IN_DAY
            self._next = _core.Timestep(final_period, next_index, delta)
        else:
            self._next = _core.Timestep(self._periods[next_index], next_index, self._deltas[next_index])

        # Return this timestep
        return current

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        if isinstance(value, pandas.Timestamp):
            self._start = value
        else:
            self._start = pandas.to_datetime(value)
        self._dirty = True

    @property
    def start_period(self):
        return pandas.Period(self.start, freq=self.freq)

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value):
        if isinstance(value, pandas.Timestamp):
            self._end = value
        else:
            self._end = pandas.to_datetime(value)
        self._dirty = True

    @property
    def end_period(self):
        return pandas.Period(self.end, freq=self.freq)

    @property
    def delta(self):
        return self._delta

    @delta.setter
    def delta(self, value):
        self._delta = value
        self._dirty = True

    @property
    def freq(self):
        d = self._delta
        if isinstance(d, int):
            freq = '{}D'.format(d)
        elif isinstance(d, datetime.timedelta):
            freq = '{}D'.format(d.days)
        else:
            freq = d
        return freq

    @property
    def offset(self):
        return pandas.tseries.frequencies.to_offset(self.freq)

    @property
    def current(self):
        """The current timestep

        If iteration has not begun this will return None.
        """
        return self._current

    @property
    def datetime_index(self):
        """ Return a `pandas.DatetimeIndex` using the start, end and delta of this object

        This is useful for creating `pandas.DataFrame` objects from Model results
        """
        return pandas.period_range(self.start, self.end, freq=self.freq)

    def __repr__(self):
        start = self.start.strftime("%Y-%m-%d")
        end = self.end.strftime("%Y-%m-%d")
        return "<Timestepper start=\"{}\" end=\"{}\" freq=\"{}\">".format(
            start, end, self.freq
        )



import numpy as np
import platypus
from . import cache_constraints, cache_objectives, cache_variable_parameters, BaseOptimisationWrapper

import logging
logger = logging.getLogger(__name__)


class PlatypusWrapper(BaseOptimisationWrapper):
    """ A helper class for running pywr optimisations with platypus.
    """
    def __init__(self, *args, **kwargs):
        super(PlatypusWrapper, self).__init__(*args, **kwargs)

        # To determine the number of variables, etc
        m = self.model

        # Cache the variables, objectives and constraints
        variables, variable_map = cache_variable_parameters(m)
        objectives = cache_objectives(m)
        constraints = cache_constraints(m)

        if len(variables) < 1:
            raise ValueError('At least one variable must be defined.')

        if len(objectives) < 1:
            raise ValueError('At least one objective must be defined.')

        self.problem = platypus.Problem(variable_map[-1], len(objectives), len(constraints))
        self.problem.function = self.evaluate
        self.problem.wrapper = self

        # Setup the problem; subclasses can change this behaviour
        self._make_variables(variables)
        self._make_constraints(constraints)

    def _make_variables(self, variables):
        """Setup the variable types. """

        ix = 0
        for var in variables:
            if var.double_size > 0:
                lower = var.get_double_lower_bounds()
                upper = var.get_double_upper_bounds()
                for i in range(var.double_size):
                    self.problem.types[ix] = platypus.Real(lower[i], upper[i])
                    ix += 1

            if var.integer_size > 0:
                lower = var.get_integer_lower_bounds()
                upper = var.get_integer_upper_bounds()
                for i in range(var.integer_size):
                    # Integers are cast to real
                    self.problem.types[ix] = platypus.Real(lower[i], upper[i])
                    ix += 1

    def _make_constraints(self, constraints):
        """ Setup the constraints. """
        # Setup the constraints
        self.problem.constraints[:] = "<=0"

    def evaluate(self, solution):
        logger.info('Evaluating solution ...')

        for ivar, var in enumerate(self.model_variables):
            j = slice(self.model_variable_map[ivar], self.model_variable_map[ivar+1])
            x = np.array(solution[j])
            assert len(x) == var.double_size + var.integer_size
            if var.double_size > 0:
                var.set_double_variables(np.array(x[:var.double_size]))

            if var.integer_size > 0:
                ints = np.round(np.array(x[-var.integer_size:])).astype(np.int32)
                var.set_integer_variables(ints)

        run_stats = self.model.run()

        objectives = []
        for r in self.model_objectives:
            sign = 1.0 if r.is_objective == 'minimise' else -1.0
            value = r.aggregated_value()
            objectives.append(sign*value)

        constraints = [r.aggregated_value() for r in self.model_constraints]

        # Return values to the solution
        logger.info('Evaluation complete!')
        if len(constraints) > 0:
            return objectives, constraints
        else:
            return objectives



from ._recorders import Recorder
import numpy as np
import pandas


class Event(object):
    """ Container for event information """
    def __init__(self, start, scenario_index):
        self.start = start
        self.scenario_index = scenario_index
        self.end = None
        self.values = None  # to record any  tracked values

    @property
    def duration(self):
        td = self.end.datetime - self.start.datetime
        return td.days


class EventRecorder(Recorder):
    """Track discrete events using a Parameter or Recorder

    The recorder works with an `IndexParameter`, `Parameter` or `Recorder`. An
    event is considered active while the value of the threshold is non-zero.

    The events are stored in a flat list across all scenarios. Each
    event is stored as a separate `Event` object. Events can be accessed as a
    dataframe using the `to_dataframe` method.

    Parameters
    ----------
    threshold - IndexParameter, Parameter or Recorder
       The object that defines the start and end of an event.
    minimum_event_length - int (default=1)
        The minimum number of time-steps that an event must last for
        to be recorded. This is useful to not record events that are
        caused by model hysteresis. The default will cause all events
        to be recorded.
    agg_func - string, callable
        Function used for aggregating across the recorders. Numpy style functions that
        support an axis argument are supported.
    event_agg_func - string, callable
        Optional different function for aggregating the `tracked_parameter` across events.
        If given this aggregation will be added as a `value` column in the `to_dataframe` method.
    tracked_parameter - `Parameter`
        The parameter to track across each event. The values from this parameter are appended each
        time-step to each event. These can then be used with other event recorders for statistical
        aggregation, or with `event_agg_func`.

     See also
     --------
     `pywr.parameters._thresholds`


     """
    def __init__(self, model, threshold, minimum_event_length=1, tracked_parameter=None, **kwargs):
        self.event_agg_func = kwargs.pop('event_agg_func', kwargs.get('agg_func'))
        super(EventRecorder, self).__init__(model, **kwargs)
        self.threshold = threshold
        self.threshold.parents.add(self)
        if minimum_event_length < 1:
            raise ValueError('Keyword "minimum_event_length" must be >= 1')
        self.minimum_event_length = minimum_event_length
        self.events = None
        self._current_events = None
        # TODO make this more generic to track components or  nodes (e.g. storage volume)
        self.tracked_parameter = tracked_parameter
        if self.tracked_parameter is not None:
            self.tracked_parameter.parents.add(self)

    def setup(self):
        pass

    def reset(self):
        self.events = []
        # This list stores if an event is current active in each scenario.
        self._current_events = [None for si in self.model.scenarios.combinations]

    def after(self):
        # Current timestep
        ts = self.model.timestepper.current

        from pywr.parameters import Parameter, IndexParameter

        if isinstance(self.threshold, Recorder):
            all_triggered = np.array(self.threshold.values(), dtype=np.int)
        elif isinstance(self.threshold, IndexParameter):
            all_triggered = self.threshold.get_all_indices()
        elif isinstance(self.threshold, Parameter):
            all_triggered = np.array(self.threshold.get_all_values(), dtype=np.int)
        else:
            raise TypeError("Threshold must be either a Recorder or Parameter instance.")

        for si in self.model.scenarios.combinations:
            # Determine if an event is active this time-step/scenario combination
            triggered = all_triggered[si.global_id]

            # Get the current event
            current_event = self._current_events[si.global_id]
            if current_event is not None:
                # A current event is active
                if triggered:
                    # Current event continues
                    # Update the timeseries of event data
                    if self.tracked_parameter is not None:
                        value = self.tracked_parameter.get_value(si)
                        current_event.values.append(value)
                else:
                    # Update the end of the current event.
                    current_event.end = ts
                    current_event.values = np.array(current_event.values)  # Convert list to nparray
                    current_length = ts.index - current_event.start.index

                    if current_length >= self.minimum_event_length:
                        # Current event ends
                        self.events.append(current_event)
                        # Event has ended; no further updates
                        current_event = None
                    else:
                        # Event wasn't long enough; don't append
                        current_event = None
            else:
                # No current event
                if triggered:
                    # Start of a new event
                    current_event = Event(ts, si)
                    # Start the timeseries of event data
                    if self.tracked_parameter is not None:
                        value = self.tracked_parameter.get_value(si)
                        current_event.values = [value, ]
                else:
                    # No event active and one hasn't started
                    # Therefore do nothing.
                    pass

            # Update list of current events
            self._current_events[si.global_id] = current_event

    def finish(self):
        ts = self.model.timestepper.current
        # Complete any unfinished events
        for si in self.model.scenarios.combinations:
            # Get the current event
            current_event = self._current_events[si.global_id]
            if current_event is not None:
                # Unfinished event
                current_event.end = ts
                self.events.append(current_event)
                self._current_events[si.global_id] = None

    def to_dataframe(self):
        """ Returns a `pandas.DataFrame` containing all of the events.

        If `event_agg_func` is a valid aggregation function and `tracked_parameter`
         is given then a "value" column is added to the dataframe containing the
         result of the aggregation.

        """
        # Return empty dataframe if no events are found.
        if len(self.events) == 0:
            return pandas.DataFrame(columns=['scenario_id', 'start', 'end'])

        scen_id = np.empty(len(self.events), dtype=np.int)
        start = np.empty_like(scen_id, dtype=object)
        end = np.empty_like(scen_id, dtype=object)
        values = np.empty_like(scen_id, dtype=float)

        for i, evt in enumerate(self.events):
            scen_id[i] = evt.scenario_index.global_id
            start[i] = evt.start.datetime
            end[i] = evt.end.datetime
            if self.tracked_parameter is not None and self.event_agg_func is not None:
                values[i] = pandas.Series(evt.values).aggregate(self.event_agg_func)

        df_dict = {'scenario_id': scen_id, 'start': start, 'end': end}
        if self.tracked_parameter is not None and self.event_agg_func is not None:
            df_dict["value"] = values

        return pandas.DataFrame(df_dict)


class EventDurationRecorder(Recorder):
    """ Recorder for the duration of events found by an EventRecorder

    This Recorder uses the results of an EventRecorder to calculate the duration
    of those events in each scenario. Aggregation by scenario is done via
    the pandas.DataFrame.groupby() method.

    Any scenario which has no events will contain a NaN value.

    Parameters
    ----------
    event_recorder : EventRecorder
        EventRecorder instance to calculate the events.
    agg_func - string, callable
        Function used for aggregating across the recorders. Numpy style functions that
        support an axis argument are supported.
    recorder_agg_func - string, callable
        Optional aggregating function for all events in each scenario. The function
        must be supported by the `DataFrame.group_by` method.

    """
    def __init__(self, model, event_recorder, **kwargs):
        # Optional different method for aggregating across self.recorders scenarios
        agg_func = kwargs.pop('recorder_agg_func', kwargs.get('agg_func'))
        self.recorder_agg_func = agg_func

        super(EventDurationRecorder, self).__init__(model, **kwargs)
        self.event_recorder = event_recorder
        self.event_recorder.parents.add(self)

    def setup(self):
        self._values = np.empty(len(self.model.scenarios.combinations))

    def reset(self):
        self._values[...] = 0.0

    def values(self):
        return self._values

    def finish(self):
        df = self.event_recorder.to_dataframe()

        self._values[...] = 0.0
        # No events found
        if len(df) == 0:
            return

        # Calculate duration
        df['duration'] = df['end'] - df['start']
        # Convert to int of days
        df['duration'] = df['duration'].dt.days
        # Drop other columns
        df = df[['scenario_id', 'duration']]

        # Group by scenario ...
        grouped = df.groupby('scenario_id').agg(self.recorder_agg_func)
        # ... and update the internal values
        for index, row in grouped.iterrows():
            self._values[index] = row['duration']


class EventStatisticRecorder(Recorder):
    """ Recorder for the duration of events found by an EventRecorder

    This Recorder uses the results of an EventRecorder to calculate aggregated statistics
    of those events in each scenario. This requires the EventRecorder to be given a `tracked_parameter`
    in order to save an array of values during each event. This recorder uses `event_agg_func` to aggregate
    those saved values in each event before applying `recorder_agg_func` to those values in each scenario.
    Aggregation by scenario is done via the pandas.DataFrame.groupby() method.

    Any scenario which has no events will contain a NaN value regardless of the aggregation function defined.

    Parameters
    ----------
    model : pywr.model.Model
    event_recorder : EventRecorder
        EventRecorder instance to calculate the events.
    agg_func - string, callable
        Function used for aggregating across the recorders. Numpy style functions that
        support an axis argument are supported.
    recorder_agg_func - string, callable
        Optional aggregating function for all events in each scenario. The function
        must be supported by the `DataFrame.group_by` method.
    event_agg_func - string, callable
        Optional different function for aggregating the `tracked_parameter` across events.
        If given this aggregation will be added as a `value` column in the `to_dataframe` method.
    """
    def __init__(self, model, event_recorder, **kwargs):
        # Optional different method for aggregating across self.recorders scenarios
        agg_func = kwargs.pop('event_agg_func', kwargs.get('agg_func'))
        self.event_agg_func = agg_func
        agg_func = kwargs.pop('recorder_agg_func', kwargs.get('agg_func'))
        self.recorder_agg_func = agg_func

        super(EventStatisticRecorder, self).__init__(model, **kwargs)
        self.event_recorder = event_recorder
        self.event_recorder.parents.add(self)

    def setup(self):
        self._values = np.empty(len(self.model.scenarios.combinations))

        if self.event_recorder.tracked_parameter is None:
            raise ValueError('To calculate event statistics requires the parent `EventRecorder` to have a `tracked_parameter`.')

    def reset(self):
        self._values[...] = np.nan

    def values(self):
        return self._values

    def finish(self):
        """ Compute the aggregated value in each scenario based on the parent `EventRecorder` events """
        events = self.event_recorder.events
        # Return NaN if no events found
        if len(events) == 0:
            return

        scen_id = np.empty(len(events), dtype=np.int)
        values = np.empty_like(scen_id, dtype=np.float64)

        for i, evt in enumerate(events):
            scen_id[i] = evt.scenario_index.global_id
            values[i] = pandas.Series(evt.values).aggregate(self.event_agg_func)

        df = pandas.DataFrame({'scenario_id': scen_id, 'value': values})

        # Group by scenario ...
        grouped = df.groupby('scenario_id').agg(self.recorder_agg_func)
        # ... and update the internal values
        for index, row in grouped.iterrows():
            self._values[index] = row['value']



from pywr._component cimport Component
from pywr._core cimport Timestep, AbstractNode, Node, AbstractStorage, Storage, ScenarioIndex, Scenario
from pywr.parameters._parameters cimport Parameter, IndexParameter


cdef class Aggregator:
    cdef object _user_func
    cdef public list func_args
    cdef public dict func_kwargs
    cdef int _func
    cpdef double aggregate_1d(self, double[:] data, ignore_nan=*) except *
    cpdef double[:] aggregate_2d(self, double[:, :] data, axis=*, ignore_nan=*) except *

cdef class Recorder(Component):
    cdef int _is_objective
    cdef public bint is_constraint
    cdef public double epsilon
    cdef public bint ignore_nan
    cdef public Aggregator _scenario_aggregator
    cpdef double aggregated_value(self) except? -1
    cpdef double[:] values(self)

cdef class AggregatedRecorder(Recorder):
    cdef object recorder_agg_func
    cdef int _recorder_agg_func
    cdef public list recorders

cdef class NodeRecorder(Recorder):
    cdef AbstractNode _node

cdef class StorageRecorder(Recorder):
    cdef AbstractStorage _node

cdef class ParameterRecorder(Recorder):
    cdef readonly Parameter _param

cdef class IndexParameterRecorder(Recorder):
    cdef readonly IndexParameter _param

cdef class NumpyArrayNodeRecorder(NodeRecorder):
    cdef Aggregator _temporal_aggregator
    cdef double[:, :] _data

cdef class NumpyArrayNodeDeficitRecorder(NumpyArrayNodeRecorder):
    pass

cdef class NumpyArrayNodeSuppliedRatioRecorder(NumpyArrayNodeRecorder):
    pass

cdef class NumpyArrayNodeCurtailmentRatioRecorder(NumpyArrayNodeRecorder):
    pass

cdef class NumpyArrayAbstractStorageRecorder(StorageRecorder):
    cdef public Aggregator _temporal_aggregator
    cdef double[:, :] _data

cdef class NumpyArrayStorageRecorder(NumpyArrayAbstractStorageRecorder):
    cdef public bint proportional

cdef class NumpyArrayLevelRecorder(NumpyArrayAbstractStorageRecorder):
    pass

cdef class NumpyArrayAreaRecorder(NumpyArrayAbstractStorageRecorder):
    pass

cdef class NumpyArrayParameterRecorder(ParameterRecorder):
    cdef public Aggregator _temporal_aggregator
    cdef double[:, :] _data

cdef class NumpyArrayIndexParameterRecorder(IndexParameterRecorder):
    cdef public Aggregator _temporal_aggregator
    cdef int[:, :] _data

cdef class FlowDurationCurveRecorder(NumpyArrayNodeRecorder):
    cdef double[:] _percentiles
    cdef double[:, :] _fdc

cdef class StorageDurationCurveRecorder(NumpyArrayStorageRecorder):
    cdef double[:] _percentiles
    cdef double[:, :] _sdc

cdef class FlowDurationCurveDeviationRecorder(FlowDurationCurveRecorder):
    cdef double[:, :] _lower_target_fdc
    cdef double[:, :] _upper_target_fdc
    cdef double[:, :] _fdc_deviations
    cdef double[:, :] _base_fdc_tile
    cdef public Scenario scenario

cdef class RollingWindowParameterRecorder(ParameterRecorder):
    cdef public Aggregator _temporal_aggregator
    cdef public int window
    cdef int position
    cdef double[:, :] _memory
    cdef double[:, :] _data

cdef class RollingMeanFlowNodeRecorder(NodeRecorder):
    cdef int position
    cdef public int timesteps
    cdef public int days
    cdef double[:, :] _memory
    cdef double[:, :] _data

cdef class BaseConstantNodeRecorder(NodeRecorder):
    cdef double[:] _values

cdef class TotalDeficitNodeRecorder(BaseConstantNodeRecorder):
    pass

cdef class TotalFlowNodeRecorder(BaseConstantNodeRecorder):
    cdef public double factor

cdef class MeanFlowNodeRecorder(BaseConstantNodeRecorder):
    cdef public double factor

cdef class DeficitFrequencyNodeRecorder(BaseConstantNodeRecorder):
    pass

cdef class BaseConstantStorageRecorder(StorageRecorder):
    cdef double[:] _values

cdef class MinimumVolumeStorageRecorder(BaseConstantStorageRecorder):
    pass

cdef class MinimumThresholdVolumeStorageRecorder(BaseConstantStorageRecorder):
    cdef public double threshold

cdef class TimestepCountIndexParameterRecorder(IndexParameterRecorder):
    cdef public int threshold
    cdef int[:] _count

cdef class AnnualCountIndexThresholdRecorder(Recorder):
    cdef public list parameters
    cdef public int threshold
    cdef int _num_years
    cdef int _ncomb
    cdef double[:, :] _data
    cdef double[:, :] _data_this_year
    cdef int _current_year
    cdef int _start_year
    cdef Aggregator _temporal_aggregator

cdef class AnnualTotalFlowRecorder(Recorder):
    cdef public list nodes
    cdef int _num_years
    cdef int _ncomb
    cdef double[:, :] _data
    cdef int _current_year
    cdef int _start_year
    cdef Aggregator _temporal_aggregator

cdef class AnnualCountIndexParameterRecorder(IndexParameterRecorder):
    cdef public int threshold
    cdef int[:] _count
    cdef int _current_year
    cdef int[:] _current_max

cdef class SeasonalFlowDurationCurveRecorder(FlowDurationCurveRecorder):
    cdef set _months


cdef class BaseConstantParameterRecorder(ParameterRecorder):
    cdef double[:] _values

cdef class TotalParameterRecorder(BaseConstantParameterRecorder):
    cdef public double factor
    cdef public bint integrate

cdef class MeanParameterRecorder(BaseConstantParameterRecorder):
    cdef public double factor



from ._parameters cimport IndexParameter, Parameter
from pywr.recorders._recorders cimport Recorder
from .._core cimport Timestep, Scenario, ScenarioIndex, AbstractNode, AbstractStorage
cimport numpy as np
ctypedef np.uint8_t uint8


cdef class AbstractThresholdParameter(IndexParameter):
    cdef public double _threshold
    cdef public Parameter _threshold_parameter
    cdef double[:] values
    cdef int predicate
    cdef public bint ratchet
    cdef uint8[:] _triggered
    cpdef double _value_to_compare(self, Timestep timestep, ScenarioIndex scenario_index) except? -1

cdef class StorageThresholdParameter(AbstractThresholdParameter):
    cdef public AbstractStorage storage

cdef class NodeThresholdParameter(AbstractThresholdParameter):
    cdef public AbstractNode node

cdef class ParameterThresholdParameter(AbstractThresholdParameter):
    cdef public Parameter param

cdef class RecorderThresholdParameter(AbstractThresholdParameter):
    cdef public Recorder recorder
    cdef public initial_value

cdef class CurrentYearThresholdParameter(AbstractThresholdParameter):
    pass

cdef class CurrentOrdinalDayThresholdParameter(AbstractThresholdParameter):
    pass




from pywr.nodes import Node, Domain, Input, Output, Link, Storage, PiecewiseLink, MultiSplitLink
from pywr.parameters import pop_kwarg_parameter, ConstantParameter, Parameter, load_parameter
from pywr.parameters.control_curves import ControlCurveParameter

DEFAULT_RIVER_DOMAIN = Domain(name='river', color='#33CCFF')

class RiverDomainMixin(object):
    def __init__(self, *args, **kwargs):
        # if 'domain' not in kwargs:
        #     kwargs['domain'] = DEFAULT_RIVER_DOMAIN
        if 'color' not in kwargs:
            self.color = '#6ECFF6' # blue
        super(RiverDomainMixin, self).__init__(*args, **kwargs)


class Catchment(RiverDomainMixin, Input):
    """A hydrological catchment, supplying water to the river network"""
    def __init__(self, *args, **kwargs):
        """Initialise a new Catchment node.

        A Catchment is an input node with a fixed inflow. I.e. min_flow and
        max_flow are the same. The value is specified as a flow keyword, and
        overrides any min_flow or max_flow keyword arguments.

        Parameters
        ----------
        flow : float or function
            The amount of water supplied by the catchment each timestep
        """
        self.color = '#82CA9D' # green
        # Grab flow from kwargs
        flow = kwargs.pop('flow', 0.0)
        # Min/max flow set in super inits
        super(Catchment, self).__init__(*args, **kwargs)
        self.flow = flow

    def get_flow(self, timestep):
        """ flow is ensured that both min_flow and max_flow are the same. """
        return self.get_min_flow(timestep)

    def __setattr__(self, name, value):
        if name == 'flow':
            self.min_flow = value
            self.max_flow = value
            return
        super(Catchment, self).__setattr__(name, value)

    @classmethod
    def load(cls, data, model):
        flow = data.pop('flow', 0.0)
        if flow is not None:
            flow = load_parameter(model, flow)
        node = super(Catchment, cls).load(data, model)
        node.flow = flow
        return node


class Reservoir(RiverDomainMixin, Storage):
    """A reservoir node with control curve.

    The Reservoir is a subclass of Storage with additional functionality to provide a
    simple control curve. The Reservoir has above_curve_cost when it is above its curve
    and the user defined cost when it is below. Typically the costs are negative
    to represent a benefit of filling the reservoir when it is below its curve.
    """
    def __init__(self, model, *args, **kwargs):
        """

        Keywords:
            control_curve - A Parameter object that can return the control curve position,
                as a percentage of fill, for the given timestep.
        """
        control_curve = pop_kwarg_parameter(kwargs, 'control_curve', None)
        above_curve_cost = kwargs.pop('above_curve_cost', None)
        cost = kwargs.pop('cost', 0.0)
        if above_curve_cost is not None:
            if control_curve is None:
                # Make a default control curve at 100% capacity
                control_curve = ConstantParameter(model, 1.0)
            elif not isinstance(control_curve, Parameter):
                # Assume parameter is some kind of constant and coerce to ConstantParameter
                control_curve = ConstantParameter(model, control_curve)

            if not isinstance(cost, Parameter):
                # In the case where an above_curve_cost is given and cost is not a Parameter
                # a default cost Parameter is created.
                kwargs['cost'] = ControlCurveParameter(model, self, control_curve, [above_curve_cost, cost])
            else:
                raise ValueError('If an above_curve_cost is given cost must not be a Parameter.')
        else:
            # reinstate the given cost parameter to pass to the parent constructors
            kwargs['cost'] = cost
        super(Reservoir, self).__init__(model, *args, **kwargs)


class River(RiverDomainMixin, Link):
    """A node in the river network

    This node may have multiple upstream nodes (i.e. a confluence) but only
    one downstream node.
    """
    def __init__(self, *args, **kwargs):
        super(River, self).__init__(*args, **kwargs)


class RiverSplit(MultiSplitLink):
    """A split in the river network

    RiverSplit is a specialised version of `pywr.nodes.MultiSplitLink` with a more convenient init method.
     It is intended for a simple case of where fixed ratio of flow is required to be distributed
     to multiple downstream routes.

    Parameters
    ----------
    factors : iterable of floats
        The factors to force on the additional splits. Number of extra_slot is assumed to be one less
        than the length of factors (as per `pywr.nodes.MultiSplitLink` documentation).
    slot_names : iterable
        The identifiers to refer to the slots when connect from this Node. Length must be one more than
         the number of extra slots required.

    See also
    --------
    pywr.nodes.MultiSplitLink

    """
    def __init__(self, *args, **kwargs):
        def _make_iterable(val):
            try:
                len(val)
            except TypeError:
                return [val]
            return val

        factors = _make_iterable(kwargs.pop('factors'))
        extra_slots = len(factors) - 1

        # These are the defaults to pass to the parent class that makes this
        # class a convenience.
        # create keyword arguments for PiecewiseLink
        kwargs['cost'] = _make_iterable(kwargs.pop('cost', 0.0))
        kwargs['max_flow'] = _make_iterable(kwargs.pop('max_flow', None))
        kwargs['extra_slots'] = extra_slots
        kwargs['factors'] = factors

        super(RiverSplit, self).__init__(*args, **kwargs)

    @classmethod
    def load(cls, data, model):
        max_flow = load_parameter(model, data.pop('max_flow', None))
        cost = load_parameter(model, data.pop('cost', 0.0))

        del(data["type"])
        return cls(model, max_flow=max_flow, cost=cost, **data)


class RiverSplitWithGauge(RiverSplit):
    """A split in the river network with a minimum residual flow

    As per `RiverSplit` but by default creates another route in the underlying object
     to model a MRF. This route is such that the MRF is not part of forced ratios. The
     intent of this object is to model the case where a proportion of flow can be
     abstracted above the MRF (e.g. 90% of flow above MRF).

    Parameters
    ----------
    mrf : float
        The minimum residual flow (MRF) at the gauge
    mrf_cost : float
        The cost of the route via the MRF
    cost : float
        The cost of the other (unconstrained) route
    factors : iterable of floats
        The factors to force on the additional splits. Number of extra_slot is assumed to be one less
        than the length of factors (as per `MultiSplitLink` documentation).
    slot_names : iterable
        The identifiers to refer to the slots when connect from this Node. Length must be one more than
         the number of extra slots required.
    """
    def __init__(self, model, name, mrf=0.0, cost=0.0, mrf_cost=0.0, **kwargs):
        kwargs['cost'] = [mrf_cost, cost]
        kwargs['max_flow'] = [mrf, None]
        super(RiverSplitWithGauge, self).__init__(model, name, **kwargs)

    @classmethod
    def load(cls, data, model):
        cost = load_parameter(model, data.pop('cost', 0.0))
        mrf_cost = load_parameter(model, data.pop('mrf_cost', 0.0))
        mrf = load_parameter(model, data.pop('mrf', 0.0))
        name = data.pop("name")
        data.pop("type", None)
        parameter = cls(model, name, mrf=mrf, cost=cost, mrf_cost=mrf_cost, **data)
        return parameter

class Discharge(Catchment):
    """An inline discharge to the river network

    This node is similar to a catchment, but sits inline to the river network,
    rather than at the head of the river.
    """
    pass

class RiverGauge(RiverDomainMixin, PiecewiseLink):
    """A river gauging station, with a minimum residual flow (MRF)
    """
    def __init__(self, *args, **kwargs):
        """Initialise a new RiverGauge instance

        Parameters
        ----------
        mrf : float
            The minimum residual flow (MRF) at the gauge
        mrf_cost : float
            The cost of the route via the MRF
        cost : float
            The cost of the other (unconstrained) route
        """
        # create keyword arguments for PiecewiseLink
        cost = kwargs.pop('cost', 0.0)
        kwargs['cost'] = [kwargs.pop('mrf_cost', 0.0), cost]
        kwargs['max_flow'] = [kwargs.pop('mrf', 0.0), None]
        super(RiverGauge, self).__init__(*args, **kwargs)

    def mrf():
        def fget(self):
            return self.sublinks[0].max_flow
        def fset(self, value):
            self.sublinks[0].max_flow = value
        return locals()
    mrf = property(**mrf())

    def mrf_cost():
        def fget(self):
            return self.sublinks[0].cost
        def fset(self, value):
            self.sublinks[0].cost = value
        return locals()
    mrf_cost = property(**mrf_cost())

    @classmethod
    def load(cls, data, model):
        mrf = load_parameter(model, data.pop("mrf"))
        mrf_cost = load_parameter(model, data.pop("mrf_cost"))
        cost = load_parameter(model, data.pop("cost", 0.0))
        del(data["type"])
        node = cls(model, mrf=mrf, mrf_cost=mrf_cost, cost=cost, **data)
        return node



import os
import json
import inspect
import numpy
from IPython.core.display import HTML, Javascript, display
from jinja2 import Template
from pywr.core import Node
from pywr.core import Model, Input, Output, Link, Storage, StorageInput, StorageOutput
from pywr.nodes import NodeMeta
from pywr._component import Component
import pywr.domains
from pywr.parameters._parameters import get_parameter_from_registry

from .figures import *

# load javascript template for d3 graph
folder = os.path.dirname(__file__)
with open(os.path.join(folder, "draw_graph.js"), "r") as f:
    draw_graph_template = Template(f.read())
with open(os.path.join(folder, "graph.css"), "r") as f:
    draw_graph_css = f.read()

def pywr_model_to_d3_json(model, attributes=False):
    """
    Convert a Pywr graph to a structure d3 can display

    Parameters
    ----------
    model : `pywr.core.Model`
    attributes: bool (default=False)
        If True, attribute data for each node is extract
    """
    nodes = []
    node_names = []
    for node in model.graph.nodes():
        if node.parent is None and node.virtual is False:
            nodes.append(node)
            node_names.append(node.name)

    edges = []
    for edge in model.graph.edges():
        node_source, node_target = edge

        # where a link is to/from a subnode, display a link to the parent instead
        if node_source.parent is not None:
            node_source = node_source.parent
        if node_target.parent is not None:
            node_target = node_target.parent

        if node_source is node_target:
            # link is between two subnodes
            continue

        index_source = node_names.index(node_source.name)
        index_target = node_names.index(node_target.name)
        edges.append({'source': index_source, 'target': index_target})

    json_nodes = []
    for n, node in enumerate(nodes):
        node_dict = {"name": node.name}
        classes = []
        cls = node.__class__
        classes.append(cls)
        while True:
            for base in cls.__bases__:
                if issubclass(base, Node) and base is not Node:
                    classes.append(base)
            if classes[-1] is cls:
                break
            else:
                cls = classes[-1]
        classes = classes[::-1]
        node_dict["clss"] = [cls.__name__.lower() for cls in classes]
        try:
            node_dict["position"] = node.position["schematic"]
        except KeyError:
            pass

        if attributes:
            node_dict["attributes"] = get_node_attr(node)

        json_nodes.append(node_dict)

    graph = {
        "nodes": json_nodes,
        "links": edges}


    return graph

def get_node_attr(node):
    """
    Returns a dictionary that contains node attributes as strings

    Parameters
    ----------
    node : a pywr node object
    """
    attrs = inspect.getmembers(node, lambda a:not(inspect.isroutine(a)))
    attribute_data = []
    for att in attrs:

        attr_name, attr_val = att
        if attr_name.startswith("_"):
            continue
        attr_type = type(attr_val).__name__

        attrs_to_skip = ["component_attrs", "components", "color", "model",  "input", "output",
                         "inputs", "outputs", "sub_domain", "sub_output", "sublinks", "visible",
                         "fully_qualified_name", "allow_isolated"]
        if not attr_val or attr_name.lower() in attrs_to_skip:
            continue

        if isinstance(attr_val, Component):
            attr_val = attr_val.name
            if not attr_val:
                attr_val = attr_type
            else:
                attr_val = attr_val + " - " + attr_type

        if isinstance(attr_val, list):
            new_vals = []
            for val in attr_val:
                val_name = str(val)
                val_name = val_name.replace("[", "").replace("]", "")
                new_vals.append(val_name)
            attr_val = "".join(new_vals)
        else:
            attr_val = str(attr_val)

        attribute_data.append({"attribute": attr_name, "value": attr_val})

    return attribute_data


def pywr_json_to_d3_json(model, attributes=False):
    """
    Converts a JSON file or a JSON-derived dict into structure that d3js can use.

    Parameters
    ----------
    model : dict or str
        str inputs should be a path to a json file containing the model.
    """

    if isinstance(model, str):
        with open(model) as d:
            model = json.load(d)

    nodes = [node["name"] for node in model["nodes"]]

    edges = []
    for edge in model["edges"]:
        sourceindex = nodes.index(edge[0])
        targetindex = nodes.index(edge[1])
        edges.append({'source': sourceindex, 'target': targetindex})

    nodes = []
    node_classes = create_node_class_trees()

    for node in model["nodes"]:
        json_node = {'name': node.pop("name"), 'clss': node_classes[node["type"].lower()]}
        try:
            json_node['position'] = node['position']['schematic']
        except KeyError:
            pass

        if attributes:
            json_node["attributes"] = []
            for name, val in node.items():

                if name == "type":
                    continue

                attr_val = val

                if isinstance(val, dict):
                    try:
                        attr_val = get_parameter_from_registry(val["type"]).__name__
                    except KeyError:
                        pass
                elif val in model["parameters"].keys():
                    param = model["parameters"][val]
                    attr_type = get_parameter_from_registry(param["type"]).__name__
                    attr_val = attr_val + " - " + attr_type

                attr_dict = {"attribute": name, "value": attr_val}
                json_node["attributes"].append(attr_dict)

        nodes.append(json_node)

    graph = {
        "nodes": nodes,
        "links": edges}

    return graph


def create_node_class_trees():
    # create class tree for each node type
    node_class_trees = {}
    for name, cls in NodeMeta.node_registry.items():
        classes = [cls]
        while True:
            for base in cls.__bases__:
                if issubclass(base, Node) and base is not Node:
                    classes.append(base)
            if classes[-1] is cls:
                break
            else:
                cls = classes[-1]
        clss = [cls.__name__.lower() for cls in classes[::-1]]
        node_class_trees[name] = clss
    return node_class_trees


def draw_graph(model, width=500, height=400, labels=False, attributes=False, css=None):
    """Display a Pywr model using D3 in Jupyter

    Parameters
    ----------
    model : pywr.core.Model or json-dict that describes a model
        The model to display
    width : int
        The width of the svg canvas to draw the graph on
    height : int
        The height of the svg canvas to draw the graph on
    labels : bool
        If True, each graph node is labelled with its name. If false, the node names are displayed
        during mouseover events
    attributes : bool
        If True, a table of node attributes is displayed during mouseover events
    css : string
        Stylesheet data to use instead of default
    """
    js = _draw_graph(model, width, height, labels, attributes, css)
    display(js)


def _draw_graph(model, width=500, height=400, labels=False, attributes=False, css=None):
    """Creates Javascript/D3 code for graph"""
    if isinstance(model, Model):
        graph = pywr_model_to_d3_json(model, attributes)
    else:
        graph = pywr_json_to_d3_json(model, attributes)

    if css is None:
        css = draw_graph_css

    js = Javascript(
        data=draw_graph_template.render(
            graph=graph,
            width=width,
            height=height,
            labels=labels,
            attributes=attributes,
            css=css.replace("\n","")
        ),
        lib="http://d3js.org/d3.v3.min.js",
    )
    return js



from .._core cimport Timestep, Scenario, ScenarioIndex, AbstractNode, Storage, AbstractStorage
from ._parameters cimport Parameter, IndexParameter

cdef class PiecewiseLinearControlCurve(Parameter):
    cdef public AbstractStorage storage_node
    cdef Parameter _control_curve
    cdef public double below_lower
    cdef public double below_upper
    cdef public double above_lower
    cdef public double above_upper
    cdef public double minimum
    cdef public double maximum

cpdef double _interpolate(double current_position, double lower_bound, double upper_bound, double lower_value, double upper_value)

cdef class BaseControlCurveParameter(Parameter):
    cdef AbstractStorage _storage_node
    cdef list _control_curves

cdef class ControlCurveInterpolatedParameter(BaseControlCurveParameter):
    cdef double[:] _values
    cdef public list parameters

cdef class ControlCurveIndexParameter(IndexParameter):
    cdef public AbstractStorage storage_node
    cdef list _control_curves


cdef class ControlCurveParameter(BaseControlCurveParameter):
    cdef double[:] _values
    cdef public list parameters
    cdef int[:] _variable_indices
    cdef double[:] _upper_bounds
    cdef double[:] _lower_bounds



"""
This module contains a Solver baseclass and several implemented subclasses.

Solvers are used to with pywr.core.Model classes to solve the network
allocation problem every time step.

Currently there are only linear programme based solvers using,
    - GLPK
    - LPSolve55
"""

solver_registry = []


class Solver(object):
    """Solver base class from which all solvers should inherit"""
    name = 'default'

    def __init__(self, *args, **kwargs):
        pass

    def setup(self, model):
        raise NotImplementedError('Solver should be subclassed to provide setup()')

    def solve(self, model):
        raise NotImplementedError('Solver should be subclassed to provide solve()')

    def reset(self):
        raise NotImplementedError('Solver should be subclassed to provide reset()')

    @property
    def stats(self):
        return {}


# Attempt to import solvers. These will only be successful if they are built correctly.
try:
    from .cython_glpk import CythonGLPKSolver as cy_CythonGLPKSolver
except ImportError:
    pass
else:
    class CythonGLPKSolver(Solver):
        """Python wrapper of Cython GLPK solver.

        This is required to subclass Solver and get the metaclass magic.
        """
        name = 'glpk'

        def __init__(self, *args, **kwargs):
            super(CythonGLPKSolver, self).__init__(*args, **kwargs)
            self._cy_solver = cy_CythonGLPKSolver(**kwargs)

        def setup(self, model):
            return self._cy_solver.setup(model)

        def solve(self, model):
            return self._cy_solver.solve(model)

        def reset(self):
            return self._cy_solver.reset()

        def dump_mps(self, filename):
            return self._cy_solver.dump_mps(filename)

        def dump_lp(self, filename):
            return self._cy_solver.dump_lp(filename)

        def dump_glpk(self, filename):
            return self._cy_solver.dump_glpk(filename)

        def retry_solve():
            def fget(self):
                return self._cy_solver.retry_solve

            def fset(self, value):
                self._cy_solver.retry_solve = value

            return locals()
        retry_solve = property(**retry_solve())

        def save_routes_flows():
            def fget(self):
                return self._cy_solver.save_routes_flows

            def fset(self, value):
                self._cy_solver.save_routes_flows = value

            return locals()
        save_routes_flows = property(**save_routes_flows())

        @property
        def routes(self):
            return self._cy_solver.routes

        @property
        def routes_flows_array(self):
            return self._cy_solver.route_flows_arr

        @property
        def stats(self):
            return self._cy_solver.stats
    solver_registry.append(CythonGLPKSolver)


try:
    from .cython_glpk import CythonGLPKEdgeSolver as cy_CythonGLPKEdgeSolver
except ImportError:
    pass
else:
    class CythonGLPKEdgeSolver(Solver):
        """Python wrapper of Cython GLPK solver.

        This is required to subclass Solver and get the metaclass magic.
        """
        name = 'glpk-edge'

        def __init__(self, *args, **kwargs):
            super(CythonGLPKEdgeSolver, self).__init__(*args, **kwargs)
            self._cy_solver = cy_CythonGLPKEdgeSolver(**kwargs)

        def setup(self, model):
            return self._cy_solver.setup(model)

        def solve(self, model):
            return self._cy_solver.solve(model)

        def reset(self):
            return self._cy_solver.reset()

        def dump_mps(self, filename):
            return self._cy_solver.dump_mps(filename)

        def dump_lp(self, filename):
            return self._cy_solver.dump_lp(filename)

        def dump_glpk(self, filename):
            return self._cy_solver.dump_glpk(filename)

        def retry_solve():
            def fget(self):
                return self._cy_solver.retry_solve

            def fset(self, value):
                self._cy_solver.retry_solve = value

            return locals()
        retry_solve = property(**retry_solve())

        @property
        def stats(self):
            return self._cy_solver.stats
    solver_registry.append(CythonGLPKEdgeSolver)


try:
    from .cython_lpsolve import CythonLPSolveSolver as cy_CythonLPSolveSolver
except ImportError:
    pass
else:
    class CythonLPSolveSolver(Solver):
        """Python wrapper of Cython LPSolve55 solver.

        This is required to subclass Solver and get the metaclass magic.
        """
        name = 'lpsolve'

        def __init__(self, *args, **kwargs):
            super(CythonLPSolveSolver, self).__init__(*args, **kwargs)
            self._cy_solver = cy_CythonLPSolveSolver(**kwargs)

        def setup(self, model):
            return self._cy_solver.setup(model)

        def solve(self, model):
            return self._cy_solver.solve(model)

        def reset(self):
            pass

        @property
        def save_routes_flows(self):
            return self._cy_solver.save_routes_flows

        @save_routes_flows.setter
        def save_routes_flows(self, value):
            self._cy_solver.save_routes_flows = value

        @property
        def routes(self):
            return self._cy_solver.routes

        @property
        def routes_flows_array(self):
            return self._cy_solver.route_flows_arr

        @property
        def stats(self):
            return self._cy_solver.stats
    solver_registry.append(CythonLPSolveSolver)



import os
import datetime
from ..parameter_property import parameter_property
from ._parameters import (
    Parameter, parameter_registry, UnutilisedDataWarning, ConstantParameter,
    ConstantScenarioParameter, ConstantScenarioIndexParameter, AnnualHarmonicSeriesParameter,
    ArrayIndexedParameter, ConstantScenarioParameter, IndexedArrayParameter,
    ArrayIndexedScenarioMonthlyFactorsParameter, TablesArrayParameter,
    DailyProfileParameter, MonthlyProfileParameter, WeeklyProfileParameter,
    ArrayIndexedScenarioParameter, ScenarioMonthlyProfileParameter, ScenarioDailyProfileParameter,
    ScenarioWeeklyProfileParameter, align_and_resample_dataframe, DataFrameParameter,
    IndexParameter, AggregatedParameter, AggregatedIndexParameter, PiecewiseIntegralParameter,
    NegativeParameter, MaxParameter, NegativeMaxParameter, MinParameter, NegativeMinParameter,
    DeficitParameter, DivisionParameter, load_parameter, load_parameter_values, load_dataframe)
from . import licenses
from ._polynomial import Polynomial1DParameter, Polynomial2DStorageParameter
from ._thresholds import StorageThresholdParameter, RecorderThresholdParameter
from ._hydropower import HydropowerTargetParameter
import numpy as np
from scipy.interpolate import interp1d
from scipy.integrate import quad
import pandas


class FunctionParameter(Parameter):
    def __init__(self, model, parent, func, *args, **kwargs):
        super(FunctionParameter, self).__init__(model, *args, **kwargs)
        self._parent = parent
        self._func = func

    def value(self, ts, scenario_index):
        return self._func(self._parent, ts, scenario_index)
FunctionParameter.register()


class ScaledProfileParameter(Parameter):
    def __init__(self, model, scale, profile, *args, **kwargs):
        super(ScaledProfileParameter, self).__init__(model, *args, **kwargs)
        self.scale = scale

        profile.parents.add(self)
        self.profile = profile

    @classmethod
    def load(cls, model, data):
        scale = float(data.pop("scale"))
        profile = load_parameter(model, data.pop("profile"))
        return cls(model, scale, profile, **data)

    def value(self, ts, si):
        p = self.profile.get_value(si)
        return self.scale * p
ScaledProfileParameter.register()


class AbstractInterpolatedParameter(Parameter):
    def __init__(self, model, x, y, interp_kwargs=None, **kwargs):
        super(AbstractInterpolatedParameter, self).__init__(model, **kwargs)
        self.x = x
        self.y = y
        self.interp = None
        default_interp_kwargs = dict(kind='linear', bounds_error=True)
        if interp_kwargs is not None:
            # Overwrite or add to defaults with given values
            default_interp_kwargs.update(interp_kwargs)
        self.interp_kwargs = default_interp_kwargs

    def _value_to_interpolate(self, ts, scenario_index):
        raise NotImplementedError()

    def setup(self):
        super(AbstractInterpolatedParameter, self).setup()
        self.interp = interp1d(self.x, self.y, **self.interp_kwargs)

    def value(self, ts, scenario_index):
        v = self._value_to_interpolate(ts, scenario_index)
        return self.interp(v)


class InterpolatedParameter(AbstractInterpolatedParameter):
    """
    Parameter value is equal to the interpolation of another parameter

    Example
    -------
    >>> x = [0, 5, 10, 20]
    >>> y = [0, 10, 30, -5]
    >>> p1 = ConstantParameter(model, 9.3) # or something more interesting
    >>> p2 = InterpolatedParameter(model, p1, x, y, interp_kwargs={"kind": "linear"})
    """
    def __init__(self, model, parameter, x, y, interp_kwargs=None, **kwargs):
        super(InterpolatedParameter, self).__init__(model, x, y, interp_kwargs, **kwargs)
        self._parameter = None
        self.parameter = parameter

    parameter = parameter_property("_parameter")

    def _value_to_interpolate(self, ts, scenario_index):
        return self._parameter.get_value(scenario_index)

    @classmethod
    def load(cls, model, data):
        parameter = load_parameter(model, data.pop("parameter"))
        x = np.array(data.pop("x"))
        y = np.array(data.pop("y"))
        kind = data.pop("kind", "linear")
        return cls(model, parameter, x, y, interp_kwargs={'kind': kind})
InterpolatedParameter.register()


class InterpolatedVolumeParameter(AbstractInterpolatedParameter):
    """
    Generic interpolation parameter calculated from current volume
    """
    def __init__(self, model, node, x, y, interp_kwargs=None, **kwargs):
        super(InterpolatedVolumeParameter, self).__init__(model, x, y, interp_kwargs, **kwargs)
        self._node = node

    def _value_to_interpolate(self, ts, scenario_index):
        return self._node.volume[scenario_index.global_id]

    @classmethod
    def load(cls, model, data):
        node = model._get_node_from_ref(model, data.pop("node"))
        volumes = np.array(data.pop("volumes"))
        values = np.array(data.pop("values"))
        kind = data.pop("kind", "linear")
        return cls(model, node, volumes, values, interp_kwargs={'kind': kind})
InterpolatedVolumeParameter.register()


class InterpolatedFlowParameter(AbstractInterpolatedParameter):
    """
    Generic interpolation parameter that uses a node's  flow at the previous time-step for interpolation.

    """
    def __init__(self, model, node, x, y, interp_kwargs=None, **kwargs):
        super().__init__(model, x, y, interp_kwargs, **kwargs)
        self._node = node

    def _value_to_interpolate(self, ts, scenario_index):       
        return self._node.prev_flow[scenario_index.global_id]

    @classmethod
    def load(cls, model, data):
        node = model._get_node_from_ref(model, data.pop("node"))
        volumes = np.array(data.pop("flows"))
        values = np.array(data.pop("values"))
        kind = data.pop("kind", "linear")
        return cls(model, node, volumes, values, interp_kwargs={'kind': kind})
InterpolatedFlowParameter.register()


class InterpolatedQuadratureParameter(AbstractInterpolatedParameter):
    """Parameter value is equal to the quadrature of the interpolation of another parameter

    Parameters
    ----------
    upper_parameter : Parameter
        Upper value of the interpolated interval to integrate over.
    x : array_like
        x coordinates of the data points for interpolation.
    y : array_like
        y coordinates of the data points for interpolation.
    lower_parameter : Parameter or None
        Lower value of the interpolated interval to integrate over. Can be `None` in which
        case the lower value of interval is zero.
    interp_kwargs : dict
        Dictionary of keyword arguments to pass to `scipy.interpolate.interp1d` class and used
        for interpolation.

    Example
    -------
    >>> x = [0, 5, 10, 20]
    >>> y = [0, 10, 30, -5]
    >>> p1 = ConstantParameter(model, 9.3) # or something more interesting
    >>> p2 = InterpolatedQuadratureParameter(model, p1, x, y, interp_kwargs={"kind": "linear"})
    """
    def __init__(self, model, upper_parameter, x, y, lower_parameter=None, interp_kwargs=None, **kwargs):
        super().__init__(model, x, y, interp_kwargs, **kwargs)
        self._upper_parameter = None
        self.upper_parameter = upper_parameter
        self._lower_parameter = None
        self.lower_parameter = lower_parameter

    upper_parameter = parameter_property("_upper_parameter")
    lower_parameter = parameter_property("_lower_parameter")

    def _value_to_interpolate(self, ts, scenario_index):
        return self._upper_parameter.get_value(scenario_index)

    def value(self, ts, scenario_index):
        a = 0
        if self._lower_parameter is not None:
            a = self._lower_parameter.get_value(scenario_index)
        b = self._value_to_interpolate(ts, scenario_index)

        cost, err = quad(self.interp, a, b)
        return cost

    @classmethod
    def load(cls, model, data):
        upper_parameter = load_parameter(model, data.pop("upper_parameter"))
        lower_parameter = load_parameter(model, data.pop("lower_parameter", None))
        x = np.array(data.pop("x"))
        y = np.array(data.pop("y"))
        kind = data.pop("kind", "linear")
        return cls(model, upper_parameter, x, y, lower_parameter=lower_parameter,
                   interp_kwargs={'kind': kind})
InterpolatedQuadratureParameter.register()


class ScenarioWrapperParameter(Parameter):
    """Parameter that utilises a different child parameter in each scenario ensemble.

    This parameter is used to switch between different child parameters based on different
    ensembles in a given `Scenario`. It can be used to vary data in a non-scenario aware
    parameter type across multiple scenario ensembles. For example, many of control curve or
    interpolation parameters do not explicitly support scenarios. This parameter can be used
    to test multiple control curve definitions as part of a single simulation.

    Parameters
    ----------
    scenario : Scenario
        The scenario instance which is used to select the parameters.
    parameters : iterable of Parameter instances
        The child parameters that are used in each of `scenario`'s ensembles. The number
        of parameters must equal the size of the given scenario.

    """
    def __init__(self, model, scenario, parameters, **kwargs):
        super().__init__(model, **kwargs)
        if scenario.size != len(parameters):
            raise ValueError("The number of parameters must equal the size of the scenario.")
        self.scenario = scenario
        self.parameters = []
        for p in parameters:
            self.children.add(p)
            self.parameters.append(p)
        # Initialise internal attributes
        self._scenario_index = None

    def setup(self):
        super().setup()
        # This setup must find out the index of self._scenario in the model
        # so that it can return the correct value in value()
        self._scenario_index = self.model.scenarios.get_scenario_index(self.scenario)

    def value(self, ts, scenario_index):
        # This is a bit confusing.
        # scenario_indices contains the current scenario number for all
        # the Scenario objects in the model run. We have cached the
        # position of self._scenario in self._scenario_index to lookup the
        # correct number to use in this instance.
        parameter = self.parameters[scenario_index.indices[self._scenario_index]]
        return parameter.get_value(scenario_index)

    @classmethod
    def load(cls, model, data):
        scenario = model.scenarios[data.pop('scenario')]

        parameters = [load_parameter(model, p) for p in data.pop('parameters')]
        return cls(model, scenario, parameters, **data)
ScenarioWrapperParameter.register()


def pop_kwarg_parameter(kwargs, key, default):
    """Pop a parameter from the keyword arguments dictionary

    Parameters
    ----------
    kwargs : dict
        A keyword arguments dictionary
    key : string
        The argument name, e.g. 'flow'
    default : object
        The default value to use if the dictionary does not have that key

    Returns a Parameter
    """
    value = kwargs.pop(key, default)
    if isinstance(value, Parameter):
        return value
    elif callable(value):
        # TODO this is broken?
        return FunctionParameter(self, value)
    else:
        return value


class PropertiesDict(dict):
    def __setitem__(self, key, value):
        if not isinstance(value, Property):
            value = ConstantParameter(value)
        dict.__setitem__(self, key, value)




from pywr.timestepper import Timestepper
from pywr.model import Model
from pywr.model import ModelStructureError, ModelDocumentWarning # TODO: exceptions module?
from pywr.nodes import *
from pywr._core import Scenario
from pywr._component import Component



from .parameters._parameters cimport Parameter


cdef class Scenario:
    cdef basestring _name
    cdef int _size
    cdef public slice slice
    cdef object _ensemble_names

cdef class ScenarioCollection:
    cdef public object model
    cdef list _scenarios
    cdef readonly list combinations
    cdef int[:, :] _user_combinations
    cpdef int get_scenario_index(self, Scenario sc) except? -1
    cpdef add_scenario(self, Scenario sc)
    cpdef int ravel_indices(self, int[:] scenario_indices) except? -1

cdef class ScenarioCombinations:
    cdef ScenarioCollection _collection

cdef class ScenarioIndex:
    cdef readonly int global_id
    cdef int[:] _indices

cdef class Timestep:
    cdef readonly object period
    cdef readonly int index
    cdef readonly double days
    cdef readonly int dayofyear
    cdef readonly int day
    cdef readonly int month
    cdef readonly int year

cdef class Domain:
    cdef object name

cdef class AbstractNode:
    cdef double[:] _prev_flow
    cdef double[:] _flow
    cdef list _recorders
    cdef Domain _domain
    cdef AbstractNode _parent
    cdef object _model
    cdef object _name
    cdef bint _allow_isolated
    cdef public bint virtual
    cdef public object __data
    cdef public basestring comment

    cdef Parameter _cost_param
    cpdef double get_cost(self, ScenarioIndex scenario_index) except? -1

    cpdef setup(self, model)
    cpdef reset(self)
    cpdef before(self, Timestep ts)
    cpdef commit(self, int scenario_index, double value)
    cpdef commit_all(self, double[:] value)
    cpdef after(self, Timestep ts)
    cpdef finish(self)
    cpdef check(self,)

cdef class Node(AbstractNode):
    cdef double _cost
    cdef double _min_flow
    cdef double _max_flow
    cdef double _conversion_factor
    cdef Parameter _min_flow_param
    cdef Parameter _max_flow_param

    cdef Parameter _conversion_factor_param
    cpdef double get_min_flow(self, ScenarioIndex scenario_index) except? -1
    cpdef double get_max_flow(self, ScenarioIndex scenario_index) except? -1
    cpdef double get_conversion_factor(self) except? -1
    cdef set_parameters(self, ScenarioIndex scenario_index)

cdef class AggregatedNode(AbstractNode):
    cdef list _nodes
    cdef double[:] _factors
    cdef double[:] _flow_weights
    cdef double _max_flow
    cdef double _min_flow
    cdef Parameter _min_flow_param
    cdef Parameter _max_flow_param

    cpdef double get_min_flow(self, ScenarioIndex scenario_index) except? -1
    cpdef double get_max_flow(self, ScenarioIndex scenario_index) except? -1

cdef class BaseInput(Node):
    cdef object _licenses

cdef class AbstractStorage(AbstractNode):
    cdef public double[:] _volume
    cdef public double[:] _current_pc

cdef class Storage(AbstractStorage):
    cdef double _cost
    cdef double _initial_volume
    cdef double _initial_volume_pc
    cdef double _min_volume
    cdef double _max_volume
    cdef double _level
    cdef double _area
    cdef Parameter _min_volume_param
    cdef Parameter _max_volume_param
    cdef Parameter _level_param
    cdef Parameter _area_param
    cpdef _reset_storage_only(self)
    cpdef double get_min_volume(self, ScenarioIndex scenario_index) except? -1
    cpdef double get_max_volume(self, ScenarioIndex scenario_index) except? -1
    cpdef double get_level(self, ScenarioIndex scenario_index) except? -1
    cpdef double get_area(self, ScenarioIndex scenario_index) except? -1

cdef class AggregatedStorage(AbstractStorage):
    cdef list _storage_nodes

cdef class VirtualStorage(Storage):
    cdef list _nodes
    cdef double[:] _factors





cdef class Component:
    cdef object _name
    cdef readonly object model
    cdef public basestring comment
    cdef readonly object parents
    cdef readonly object children
    cpdef setup(self)
    cpdef reset(self)
    cpdef before(self)
    cpdef after(self)
    cpdef finish(self)


from pywr.recorders._recorders cimport Recorder
from pywr._component cimport Component
from .._core cimport Timestep, Scenario, ScenarioIndex, ScenarioCollection, AbstractNode, Node

cdef class Parameter(Component):
    cdef public int double_size
    cdef public int integer_size
    cdef int _size
    cdef public bint is_variable
    cdef AbstractNode _node
    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1
    cdef double[:] __values
    cdef calc_values(self, Timestep ts)
    cpdef double get_value(self, ScenarioIndex scenario_index)
    cpdef double[:] get_all_values(self)

    # New variable API
    cpdef set_double_variables(self, double[:] values)
    cpdef double[:] get_double_variables(self)
    cpdef double[:] get_double_lower_bounds(self)
    cpdef double[:] get_double_upper_bounds(self)
    cpdef set_integer_variables(self, int[:] values)
    cpdef int[:] get_integer_variables(self)
    cpdef int[:] get_integer_lower_bounds(self)
    cpdef int[:] get_integer_upper_bounds(self)



cdef class ConstantParameter(Parameter):
    cdef double _value
    cdef public double scale, offset
    cdef double[:] _lower_bounds
    cdef double[:] _upper_bounds

cdef class DataFrameParameter(Parameter):
    cdef double[:,:] _values
    cdef public Scenario scenario
    cdef int _scenario_index
    cdef public object dataframe

cdef class ArrayIndexedParameter(Parameter):
    cdef double[:] values

cdef class ArrayIndexedScenarioParameter(Parameter):
    cdef Scenario _scenario
    cdef double[:, :] values
    cdef int _scenario_index

cdef class ConstantScenarioParameter(Parameter):
    cdef Scenario _scenario
    cdef double[:] _values
    cdef int _scenario_index

cdef class ArrayIndexedScenarioMonthlyFactorsParameter(Parameter):
    cdef double[:] _values
    cdef double[:, :] _factors
    cdef Scenario _scenario
    cdef int _scenario_index



cdef class DailyProfileParameter(Parameter):
    cdef double[:] _values

cdef class WeeklyProfileParameter(Parameter):
    cdef double[:] _values

cdef class MonthlyProfileParameter(Parameter):
    cdef double[:] _values
    cdef double[:] _interp_values
    cdef double[:] _lower_bounds
    cdef double[:] _upper_bounds
    cdef public object interp_day
    cpdef _interpolate(self)

cdef class ScenarioMonthlyProfileParameter(Parameter):
    cdef double[:, :] _values
    cdef Scenario _scenario
    cdef int _scenario_index

cdef class ScenarioDailyProfileParameter(Parameter):
    cdef double[:, :] _values
    cdef Scenario _scenario
    cdef int _scenario_index

cdef class ScenarioWeeklyProfileParameter(Parameter):
    cdef double[:, :] _values
    cdef Scenario _scenario
    cdef int _scenario_index

cdef class IndexParameter(Parameter):
    cpdef int index(self, Timestep timestep, ScenarioIndex scenario_index) except? -1
    cdef int[:] __indices
    cpdef int get_index(self, ScenarioIndex scenario_index)
    cpdef int[:] get_all_indices(self)

cdef class ConstantScenarioIndexParameter(IndexParameter):
    cdef Scenario _scenario
    cdef int[:] _values
    cdef int _scenario_index

cdef class TablesArrayParameter(IndexParameter):
    cdef double[:, :] _values_dbl
    cdef int[:, :] _values_int
    cdef public Scenario scenario
    cdef public object h5file
    cdef public object h5store
    cdef public object node
    cdef public object where

    cdef int _scenario_index
    cdef int[:] _scenario_ids

cdef class IndexedArrayParameter(Parameter):
    cdef public IndexParameter index_parameter
    cdef public list params

cdef class AnnualHarmonicSeriesParameter(Parameter):
    cdef public double mean
    cdef double[:] _amplitudes, _phases
    cdef double _mean_lower_bounds, _mean_upper_bounds
    cdef double[:] _amplitude_lower_bounds, _amplitude_upper_bounds
    cdef double[:] _phase_lower_bounds, _phase_upper_bounds
    cdef double _value_cache
    cdef int _ts_index_cache

cdef class AggregatedParameter(Parameter):
    # This is a list rather than a set due to floating point arithmetic.
    # The order is important for maintaining determinism.
    cdef public list parameters
    cdef object _agg_user_func
    cdef int _agg_func
    cpdef double value(self, Timestep timestep, ScenarioIndex scenario_index) except? -1
    cpdef add(self, Parameter parameter)
    cpdef remove(self, Parameter parameter)

cdef class AggregatedIndexParameter(IndexParameter):
    # This is a list; see above AggregatedParameter.
    cdef public list parameters
    cdef object _agg_user_func
    cdef int _agg_func
    cpdef double value(self, Timestep timestep, ScenarioIndex scenario_index) except? -1
    cpdef int index(self, Timestep timestep, ScenarioIndex scenario_index) except? -1
    cpdef add(self, Parameter parameter)
    cpdef remove(self, Parameter parameter)


cdef class DivisionParameter(Parameter):
    cdef Parameter _numerator
    cdef Parameter _denominator


cdef class NegativeParameter(Parameter):
    cdef public Parameter parameter

cdef class MaxParameter(Parameter):
    cdef public Parameter parameter
    cdef public double threshold

cdef class NegativeMaxParameter(MaxParameter):
    pass

cdef class MinParameter(Parameter):
    cdef public Parameter parameter
    cdef public double threshold

cdef class NegativeMinParameter(MinParameter):
    pass

cdef class DeficitParameter(Parameter):
    cdef public Node node

cdef class FlowParameter(Parameter):
    cdef public Node node
    cdef double[:] __next_values
    cdef public double initial_value

cdef class PiecewiseIntegralParameter(Parameter):
    cdef public double[:] x
    cdef public double[:] y
    cdef public Parameter parameter
